"""
Convenience function for getting info about the current OS platform
"""
import os
import sys

import torch


def get_device_type():
    """returns the device type as a string"""
    #return 'cpu'
    # noinspection PyUnresolvedReferences
    if torch.cuda.is_available():
        return 'cuda'
    elif torch.backends.mps.is_available():
        return 'mps'
    else:
        return 'cpu'


def get_cuda_device_count():
    return torch.cuda.device_count()


def get_cpu_device_count():
    return os.cpu_count()


def is_linux():
    return sys.platform == 'linux' or sys.platform == 'linux2'


def is_mac():
    return sys.platform == 'darwin'


def is_windows():
    return sys.platform == 'win32'


def expand_user_home_path(path, user, platform):
    if platform == 'darwin':
        path = path.replace('~/', f'/Users/{user}/')
    elif platform == 'windows':
        path = path.replace('~/', f'C:/Users/{user}/')
    else:
        path = path.replace('~/', f'/home/{user}/')
    return path



from ray import logger


def inspect_ray_serializability(obj):
    """
    Check serializability of an object using Ray's inspect_serializability function.
    Please note that there is currently an issue with pickling when the multiprocessing module is imported
    """
    from ray.util import inspect_serializability
    for attribute, value in obj.__dict__.items():
        logger.info(f'inspecting serializability of {attribute}...')
        inspect_serializability(value)



import shutil
import subprocess
import sys
import time
from typing import Optional, List

from ray import logger
from ray.tune.syncer import Syncer, DEFAULT_SYNC_PERIOD

from util.platform import expand_user_home_path


class RsyncSyncer(Syncer):

    def __init__(self,
                 driver_user,
                 driver_hostname,
                 driver_ssh_port,
                 worker_user,
                 worker_hostname,
                 worker_ssh_port,
                 worker_platform,
                 use_sync_period=False,
                 sync_at_train_end_only=True,
                 sync_period=DEFAULT_SYNC_PERIOD):
        super().__init__(sync_period)
        self.driver_user = driver_user
        self.driver_hostname = driver_hostname
        self.driver_ssh_port = driver_ssh_port
        self.driver_platform = sys.platform
        self.worker_user = worker_user
        self.worker_hostname = worker_hostname
        self.worker_ssh_port = worker_ssh_port
        self.worker_platform = worker_platform
        self.use_sync_period = use_sync_period
        self.sync_at_train_end_only = sync_at_train_end_only
        self.training_has_ended = False
        self.is_syncing = False

    def sync_up(self, local_dir: str, remote_dir: str, exclude: Optional[List] = None) -> bool:
        """Synchronize the local directory (Ray worker) to the remote directory (driver)."""
        try:
            if self.sync_at_train_end_only and not self.training_has_ended:
                return False
            if self.is_syncing:
                return False
            self.is_syncing = True
            if remote_dir.startswith('file://'):
                remote_dir = remote_dir[7:]
            remote_dir = expand_user_home_path(remote_dir, self.driver_user, self.driver_platform)
            sync_cmd = f'rsync -avz -e "ssh -i ~/.ssh/id_rsa -o StrictHostKeyChecking=no" -p {self.driver_ssh_port} {local_dir}/ {self.driver_user}@{self.driver_hostname}:{remote_dir}/'
            logger.info(f'syncing from ray worker to local computer: {sync_cmd}')
            try:
                output = str(subprocess.check_output(sync_cmd, shell=True)).replace('\\n', '\n')
                logger.debug(output)
                self.last_sync_up_time = time.time()
                return True
            except subprocess.CalledProcessError as e:
                logger.error(f'error syncing up: {e}')
                return False
        finally:
            self.is_syncing = False

    def sync_down(self, remote_dir: str, local_dir: str, exclude: Optional[List] = None) -> bool:
        """Synchronize the remote directory (driver) to the local directory (Ray worker)."""
        try:
            if self.is_syncing:
                return False
            self.is_syncing = True
            if remote_dir.startswith('file://'):
                remote_dir = remote_dir[7:]
            remote_dir = expand_user_home_path(remote_dir, self.driver_user, self.driver_platform)
            sync_cmd = f'rsync -avz -e "ssh -i ~/.ssh/id_rsa -o StrictHostKeyChecking=no" --delete --ignore-errors -p {self.worker_ssh_port} {remote_dir}/ {self.worker_user}@{self.worker_hostname}:{local_dir}/'
            logger.info(f'syncing from local computer to ray worker: {sync_cmd}')
            try:
                output = str(subprocess.check_output(sync_cmd, shell=True)).replace('\\n', '\n')
                #logger.debug(output)
                self.last_sync_down_time = time.time()
                return True
            except subprocess.CalledProcessError as e:
                logger.error(f'error syncing down: {e}')
                return False
        finally:
            self.is_syncing = False

    def sync_up_if_needed(self, local_dir: str, remote_dir: str, exclude: Optional[List] = None):
        if self.use_sync_period:
            return super().sync_up_if_needed(local_dir, remote_dir, exclude)
        elif self.training_has_ended:
            return super().sync_up(local_dir, remote_dir, exclude)
        return False

    def sync_down_if_needed(self, remote_dir: str, local_dir: str, exclude: Optional[List]           = None):
        if self.use_sync_period:
            return super().sync_down_if_needed(remote_dir, local_dir, exclude)
        return False

    def delete(self, local_dir: str) -> bool:
        if self.training_has_ended:
            return False
        if self.is_syncing:
            logger.warn(f'last sync still in progress, skipping deletion of {local_dir}')
            return False
        if local_dir.startswith('file://'):
            local_dir = local_dir[7:]
        worker_dir = expand_user_home_path(local_dir, self.worker_user, self.worker_platform)
        logger.info(f'deleting {worker_dir}')
        shutil.rmtree(worker_dir)
        return True



#!/usr/bin/env python
import os
from distutils.command.sdist import sdist

from setuptools import setup, Command


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./src/*.egg-info')


if __name__ == "__main__":
    setup(
        cmdclass={
            'clean': CleanCommand,
        }
    )



"""
Utilities for working with Ray.
"""
import os
import shutil
import subprocess

import ray
import yaml
from ray import logger
from ray_quickstart.rsync_syncer import RsyncSyncer
from ray_quickstart.util import platform
from ray_quickstart.util.platform import expand_user_home_path


def initialize_ray_with_syncer(base_dir, src_dir, config_file_path, success_callback=None):
    if ray.is_initialized():
        return
    if not os.path.exists(config_file_path):
        raise FileNotFoundError(f'config file not found at {config_file_path}')
    with open(config_file_path, 'r') as f:
        props = yaml.safe_load(f)
    driver_user = props['driver']['user']
    driver_hostname_or_ip_address = props['driver']['hostname_or_ip_address']
    driver_ssh_port = props['driver']['ssh_port']
    worker_user = props['worker']['user']
    worker_hostname_or_ip_address = props['worker']['hostname_or_ip_address']
    worker_ssh_port = props['worker']['ssh_port']
    syncer = RsyncSyncer(driver_user,
                         driver_hostname_or_ip_address,
                         driver_ssh_port,
                         worker_user,
                         worker_hostname_or_ip_address,
                         worker_ssh_port,
                         'linux')
    try:
        clear_local_trial_results_dir(syncer)
        _do_initialize_ray(src_dir, props, success_callback)
        return syncer
    except ConnectionError:
        if platform.is_windows():
            with subprocess.Popen(f'{base_dir}/scripts/ray_start.bat') as p:
                p.wait()
            _do_initialize_ray(src_dir, props, success_callback)
            return syncer
        else:
            raise


def _do_initialize_ray(src_dir, props, success_callback=None):
    # runtime_env is required for cloudpickle to be able to find modules
    runtime_env = {'working_dir': src_dir}
    #ray.init(local_mode=True)
    ray_head_hostname_or_ip_address = props['ray_head']['head_hostname_or_ip_address']
    ray_head_client_server_port = props['ray_head']['client_server_port']
    ray.init(address=f'ray://{ray_head_hostname_or_ip_address}:{ray_head_client_server_port}', runtime_env=runtime_env)
    if success_callback is not None:
        success_callback()



def get_local_trial_results_dir():
    return f'~/ray_results'


def clear_local_trial_results_dir(syncer):
    logger.info('clearing local trial results dir...')
    local_trial_results_dir = os.path.expanduser(get_local_trial_results_dir())
    if os.path.exists(local_trial_results_dir):
        delete_dir_contents(local_trial_results_dir)
    else:
        os.makedirs(local_trial_results_dir, exist_ok=True)
    if syncer is not None:
        trial_results_dir = get_local_trial_results_dir()
        syncer.sync_down(trial_results_dir,
                         expand_user_home_path(trial_results_dir,
                                               syncer.worker_user,
                                               syncer.worker_platform))


def delete_dir_contents(dir_path):
    for filename in os.listdir(dir_path):
        file_path = os.path.join(dir_path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path, ignore_errors=True)
